from watsor.config.coco import COCO_CLASSES
from watsor.stream.share import Detection


class ConfidenceFilter(object):
    """Filters only detections with a confidence greater than the threshold set.
    Threshold is set per label (detection object class).
    """

    def __init__(self, camera_config):
        self.__indexes = {}
        for entry in camera_config['detect']:
            coco_class = next(iter(entry))
            idx = COCO_CLASSES.index(coco_class)
            self.__indexes[idx] = entry[coco_class]['confidence'] / 100

    def __call__(self, detection: Detection):
        confidence = self.__indexes.get(detection.label, None)
        return confidence is not None and detection.confidence >= confidence



import numpy as np
from unittest import TestCase, installHandler, main
from unittest.mock import MagicMock
from threading import Thread
from logging import getLogger
from logging.handlers import QueueHandler
from multiprocessing import Queue, Event
from time import time
from watsor.stream.share import FrameBuffer, State
from watsor.output.snapshot import Snapshot
from watsor.output.draw import DrawEffect
from watsor.stream.log import LogHandler
from watsor.stream.work import Payload
from watsor.config.coco import COCO_CLASSES


class MockEffect(object):

    def apply(self, image_in, image_out, shape, header_in, header_out):
        for detection in filter(lambda d: d.label > 0, header_out.detections):
            self.draw_rect(detection.bounding_box.x_min, detection.bounding_box.y_min,
                           detection.bounding_box.x_max, detection.bounding_box.y_max)

    @staticmethod
    def draw_rect(x1, y1, x2, y2):
        pass


class TestOutput(TestCase):

    def setUp(self):
        """Handle Ctrl+C signal to stop tests prematurely.
        """

        installHandler()

    def test_draw(self):
        width = 2
        height = 2

        frame_buffer = FrameBuffer(1, width, height)
        frame = frame_buffer.frames[0]

        self._mock_detections(frame)

        image_shape, image_np_in = frame.get_numpy_image(np.uint8)
        image_np_out = np.array(image_np_in)

        effect = DrawEffect()
        effect.apply(image_np_in, image_np_out, image_shape, frame.header, frame.header)

    def test_snapshot(self):
        width = 100
        height = 100
        frame_buffer = FrameBuffer(10, width, height)
        frame_queue = Queue(1)

        log_queue = Queue()
        getLogger().addHandler(QueueHandler(log_queue))

        stop_process_event = Event()

        effect = MockEffect()
        effect.draw_rect = MagicMock()

        snapshot = Snapshot("snapshot", stop_process_event, log_queue, frame_queue, frame_buffer,
                            self._create_detect_config(width, height), [effect])
        processes = [snapshot,
                     LogHandler(Thread, "logger", stop_process_event, log_queue, filename=None)]

        for process in processes:
            process.start()

        try:
            frame_index = 0
            frame = frame_buffer.frames[frame_index]

            self._mock_detections(frame)

            frame.latch.next()
            frame.latch.next()

            payload = Payload(None, frame_index)
            frame_queue.put(payload)

            self.assertTrue(frame.latch.wait_for(State.READY, stop_process_event.is_set, 10))

            with self.assertRaises(AssertionError):
                snapshot.get('person')

            self.assertIsNotNone(snapshot.get('book'))

            effect.draw_rect.assert_called_with(1, 2, 3, 4)
        finally:
            stop_process_event.set()
            for process in processes:
                process.join(30)

    @staticmethod
    def _create_detect_config(width, height):
        return {
            'width': width,
            'height': height,
            'detect': [{'book': {}}]
        }

    @staticmethod
    def _mock_detections(frame):
        frame.header.detections[0].label = COCO_CLASSES.index('book')
        frame.header.detections[0].bounding_box.x_min = 1
        frame.header.detections[0].bounding_box.y_min = 2
        frame.header.detections[0].bounding_box.x_max = 3
        frame.header.detections[0].bounding_box.y_max = 4
        frame.header.epoch = time()


if __name__ == '__main__':
    main(verbosity=2)



import os
from unittest import TestCase, main
from PIL import Image, ImageDraw
from tempfile import NamedTemporaryFile
from watsor.filter.confidence import ConfidenceFilter
from watsor.filter.area import AreaFilter
from watsor.filter.mask import MaskFilter
from watsor.filter.track import TrackFilter
from watsor.stream.share import Detection, BoundingBox


class TestFilter(TestCase):

    def test_confidence(self):
        confidence_filter = ConfidenceFilter({'detect': [{
            'person': {
                'confidence': 50
            }}]})

        self.assertTrue(confidence_filter(Detection(label=1, confidence=0.70)))
        self.assertFalse(confidence_filter(Detection(label=1, confidence=0.40)))
        self.assertFalse(confidence_filter(Detection(label=2, confidence=0.70)))

    def test_area(self):
        area_filter = AreaFilter({
            'width': 100,
            'height': 100,
            'detect': [{
                'person': {
                    'area': 50
                }}]
        })

        self.assertTrue(area_filter(Detection(label=1, confidence=0.70, bounding_box=BoundingBox(0, 0, 100, 50))))
        self.assertFalse(area_filter(Detection(label=1, confidence=0.70, bounding_box=BoundingBox(0, 0, 50, 50))))
        self.assertFalse(area_filter(Detection(label=2, confidence=0.70, bounding_box=BoundingBox(0, 0, 100, 50))))

    def test_mask(self):
        with self.assertRaisesRegex(AssertionError, "Error reading mask file"):
            MaskFilter({'width': 1, 'height': 1, 'mask': 'notafile.png'})

        tmp_mask_file = NamedTemporaryFile(suffix='.png', delete=False)
        try:
            with Image.new('RGB', (10, 10)) as image:
                image.save(tmp_mask_file.name)

            with self.assertRaisesRegex(AssertionError, "Mask image .+ is not of 32 bit color"):
                MaskFilter({'width': 10, 'height': 10, 'mask': tmp_mask_file.name})

            with Image.new('RGBA', (100, 100)) as image:
                with Image.new("L", image.size) as alpha:
                    draw = ImageDraw.Draw(alpha)
                    draw.rectangle((50, 0, alpha.width, alpha.height), fill=255)
                    image.putalpha(alpha)
                image.save(tmp_mask_file.name)

            with self.assertRaisesRegex(AssertionError, "The size of mask image .+ doesn't match"):
                MaskFilter({'width': 50, 'height': 50, 'mask': tmp_mask_file.name})

            mask_filter = MaskFilter({
                'width': image.width,
                'height': image.height,
                'mask': tmp_mask_file.name,
                'detect': []
            })
        finally:
            # Delete explicitly for compatibility with Windows NT
            tmp_mask_file.close()
            os.unlink(tmp_mask_file.name)

        self.assertFalse(mask_filter(Detection(label=1, confidence=0.70, bounding_box=BoundingBox(20, 20, 40, 80))))
        detection = Detection(label=1, confidence=0.70, bounding_box=BoundingBox(20, 20, 80, 80))
        self.assertTrue(mask_filter(detection))
        self.assertEqual(1, detection.zones[0])

    def test_track(self):
        track_filter = TrackFilter(sensitivity=1, history=2)
        detections, suspicious_activity = track_filter([
            Detection(label=1, confidence=0.70, bounding_box=BoundingBox(50, 50, 60, 60)),
            Detection(label=1, confidence=0.70, bounding_box=BoundingBox(10, 10, 30, 30))
        ])

        self.assertTrue(suspicious_activity)
        self.assertEqual(2, len(detections))
        self.assertListEqual([50, 50, 60, 60], self._to_array(detections[0].bounding_box))
        self.assertListEqual([10, 10, 30, 30], self._to_array(detections[1].bounding_box))

        detections, suspicious_activity = track_filter([
            Detection(label=1, confidence=0.70, bounding_box=BoundingBox(40, 40, 55, 55)),
            Detection(label=1, confidence=0.70, bounding_box=BoundingBox(80, 80, 90, 90))
        ])

        self.assertTrue(suspicious_activity)
        self.assertEqual(2, len(detections))
        self.assertListEqual([40, 40, 60, 60], self._to_array(detections[0].bounding_box))
        self.assertListEqual([80, 80, 90, 90], self._to_array(detections[1].bounding_box))

    @staticmethod
    def _to_array(bounding_box: BoundingBox):
        return [bounding_box.x_min, bounding_box.y_min, bounding_box.x_max, bounding_box.y_max]


if __name__ == '__main__':
    main(verbosity=2)



import traceback
import logging
import logging.handlers
from sys import stdout, stderr
from watsor.stream.work import Work


class LogHandler(Work):
    """Performs the logging in a console and a file, receiving the messages from multiple processes
    through the queue. While the normal messages are being written to the console, teh errors are being
    reported to stderr. The log file is being rolled over after the current file reaches a certain size.
    """

    def __init__(self, delegate_class, name: str, stop_event, log_queue, filename,
                 max_bytes=10 * 1024 * 1024, backup_count=5, kwargs=None):
        super().__init__(delegate_class, name, stop_event, log_queue, log_queue,
                         args=(filename, max_bytes, backup_count),
                         kwargs={} if kwargs is None else kwargs)

    def _config_logger(self, log_queue, *args, **kwargs):
        self._config_log_handers(*args, **kwargs)

    def _config_log_handers(self, _, filename, max_bytes, backup_count, *args, **kwargs):
        self._logger = logging.Logger(self.__class__.__name__)

        console_stdout = logging.StreamHandler(stdout)
        formatter = logging.Formatter('%(threadName)-16s %(name)-24s %(levelname)-8s: %(message)s')
        console_stdout.addFilter(lambda record: record.levelno < logging.ERROR)
        console_stdout.setFormatter(formatter)

        console_stderr = logging.StreamHandler(stderr)
        console_stderr.setLevel(logging.ERROR)
        console_stderr.setFormatter(formatter)

        self._logger.addHandler(console_stdout)
        self._logger.addHandler(console_stderr)

        if filename is not None:
            file = logging.handlers.RotatingFileHandler(filename, 'a', max_bytes, backup_count)
            formatter = logging.Formatter('%(asctime)s %(threadName)-16s %(name)-24s %(levelname)-8s: %(message)s')
            file.setFormatter(formatter)
            if 'DEBUG' == kwargs.get('log_level'):
                self._logger.debug("Log is being written to %s", filename)
            self._logger.addHandler(file)

    def _next_frame(self, record, *args, **kwargs):
        try:
            self._logger.handle(record)
        except Exception:
            traceback.print_exc()



import unittest
import os
from multiprocessing import set_start_method

if __name__ == '__main__':
    # Spawns the processes forcing the system to create copies of stream objects
    # in the scope of a process to validate the data and state are shared properly.
    set_start_method('spawn')

    suite = unittest.defaultTestLoader.discover(start_dir=os.path.dirname(__file__))
    result = unittest.TextTestRunner(failfast=True, descriptions=False, verbosity=2).run(suite)
    if len(result.errors) > 0 or len(result.failures) > 0:
        exit(1)



from queue import Full
from collections import Counter


class Publish(object):
    """This class keeps the list of subscribers and publishes a message only to
    those who can consume it. Thus if a subscriber is busy processing previous
    frame, it merely misses the current one. This logic ensures a slow worker
    won't make others to wait.
    """

    def __init__(self, lock):
        self._subscribers = Counter()
        self._subscribers_lock = lock

    def subscribe(self, a_queue):
        self._subscribers_lock.acquire()
        try:
            self._subscribers[a_queue] += 1
        finally:
            self._subscribers_lock.release()

    def unsubscribe(self, a_queue):
        self._subscribers_lock.acquire()
        try:
            self._subscribers[a_queue] -= 1
            if self._subscribers[a_queue] == 0:
                del self._subscribers[a_queue]
        finally:
            self._subscribers_lock.release()

    def _publish(self, payload):
        self._subscribers_lock.acquire()
        try:
            count = 0
            for a_queue in self._subscribers.keys():
                try:
                    a_queue.put_nowait(payload)
                    count += 1
                except Full:
                    pass
            return count
        finally:
            self._subscribers_lock.release()



import numpy as np
from ctypes import memmove, addressof, sizeof


class CopyHeaderEffect(object):

    @staticmethod
    def apply(image_in, image_out, shape, header_in, header_out):
        memmove(addressof(header_out.get_obj()), addressof(header_in.get_obj()), sizeof(header_in.get_obj()))


class CopyImageEffect(object):

    @staticmethod
    def apply(image_in, image_out, shape, header_in, header_out):
        np.copyto(image_out, image_in)



import subprocess as sp
import threading
import logging
import io
import signal
from time import time
from watsor.stream.read import Read, ReadPublish, ReadDetectPublish
from watsor.stream.work import Work, Payload
from watsor.stream.share import Frame, FrameBuffer, FramesPerSecond, RateLimiter

try:
    SIGSTOP = signal.SIGSTOP
    SIGCONT = signal.SIGCONT
except AttributeError:  # Windows doesn't have those signals
    SIGSTOP = None
    SIGCONT = None


class FFmpegDecoder(ReadDetectPublish):
    """Controls FFmpeg subprocess (which decodes video stream), reading raw 24-bit frames and
    putting them in shared buffer. Triggers detection and allows other workers to subscribe
    to perform their work for when detection and filtering are complete.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer, cmd_args, cwd=None, stdin=sp.DEVNULL,
                 kwargs=None):
        self.__cmd_args = cmd_args
        self.__cwd = cwd
        self.__stdin = stdin
        self.__subprocess = None
        self.__stderr_thread = None
        self.__fps = FramesPerSecond()
        self.__rate_limiter = RateLimiter()
        super().__init__(name, stop_event, log_queue, frame_queue, frame_buffer,
                         args=(self.__fps, self.__rate_limiter), kwargs={} if kwargs is None else kwargs)

    def initialize(self):
        assert self.__subprocess is None or self.__subprocess.poll() is not None, \
            "Subprocess has not terminated yet"

        self.__subprocess = sp.Popen(args=self.__cmd_args, cwd=self.__cwd,
                                     stdout=sp.PIPE, stderr=sp.PIPE, stdin=self.__stdin)
        if SIGSTOP is not None:
            self.__subprocess.send_signal(signal.SIGSTOP)

        self.__stderr_thread = threading.Thread(name=self.name, target=_stderr_reader,
                                                args=(self.__class__.__name__, self.__subprocess.stderr))
        self.__stderr_thread.daemon = True

        super().initialize()

    @property
    def stdin(self):
        return self.__subprocess.stdin

    @property
    def fps(self):
        return self.__fps

    @property
    def rate_limiter(self):
        return self.__rate_limiter

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super(Read, self)._run(stop_event, log_queue, *args, **kwargs)
        self._logger.debug(self.__cmd_args)
        self._logger.debug("Reading from stdout")
        try:
            self._spin(self._process, stop_event, *args, **kwargs)
        except EOFError:
            pass  # Ignore when end of stream (file) reached and stop gracefully
        except Exception:
            self._logger.exception('Spin failure')
        finally:
            self.__subprocess.stdout.close()
            self._logger.debug("Stdout closed")

    def _new_frame(self, frame: Frame, frame_queue, frame_buffer, fps, rate_limiter, *args, **kwargs):
        frame.clear()
        image = self.__subprocess.stdout.readinto(frame.image.get_obj())
        if not image:
            raise EOFError
        frame.header.epoch = time()

        result = rate_limiter.allow()
        if result:
            fps(value=True)
        return result

    def start(self):
        super().start()
        self.__stderr_thread.start()
        if SIGCONT is not None:
            self.__subprocess.send_signal(signal.SIGCONT)

    def terminate(self):
        if SIGCONT is not None:
            self.__subprocess.send_signal(signal.SIGCONT)
        self.__subprocess.terminate()
        super().terminate()

    def join(self, timeout=None):
        try:
            super().join(timeout)
            self.__subprocess.wait(timeout)
        except Exception:
            self.__subprocess.terminate()
            raise
        finally:
            self.__subprocess.stderr.close()


class FFmpegEncoder(Work):
    """Controls FFmpeg subprocess feeding it with raw 24-bit frames to encode them
     in compressed video stream. Exposes FFmpeg subprocess stdout for further re-streaming.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer, cmd_args, cwd=None, stdout=sp.DEVNULL,
                 args=(), kwargs=None):
        self.__cmd_args = cmd_args
        self.__cwd = cwd
        self.__stdout = stdout
        self.__subprocess = None
        self.__stderr_thread = None
        self.__fps = FramesPerSecond()
        self.__written = None
        super().__init__(threading.Thread, name, stop_event, log_queue, frame_queue,
                         args=(stop_event, frame_buffer, self.__fps, *args),
                         kwargs={} if kwargs is None else kwargs)

    def initialize(self):
        assert self.__subprocess is None or self.__subprocess.poll() is not None, \
            "Subprocess has not terminated yet"

        self.__subprocess = sp.Popen(args=self.__cmd_args, cwd=self.__cwd,
                                     stdout=self.__stdout, stderr=sp.PIPE, stdin=sp.PIPE)
        if SIGSTOP is not None:
            self.__subprocess.send_signal(signal.SIGSTOP)

        self.__stderr_thread = threading.Thread(name=self.name, target=_stderr_reader,
                                                args=(self.__class__.__name__, self.__subprocess.stderr))
        self.__stderr_thread.daemon = False

        super().initialize()

    @property
    def stdout(self):
        return self.__subprocess.stdout

    @property
    def fps(self):
        return self.__fps

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super(Work, self)._run(stop_event, log_queue, *args, **kwargs)
        self._logger.debug(self.__cmd_args)
        self._logger.debug("Writing to stdin")
        try:
            self._spin(self._process, stop_event, *args, **kwargs)
        except BrokenPipeError:
            pass  # Ignore broken pipe errors, as ffmpeg exited before all data were written.
        except Exception:
            self._logger.exception('Spin failure')
        finally:
            self._close_stdin()
            self._logger.debug("Stdin closed")

    def _next_frame(self, payload: Payload, stop_event, frame_buffer: FrameBuffer, fps, *args, **kwargs):
        frame = frame_buffer.frames[payload.frame_index]
        try:
            self.__written = self.__subprocess.stdin.write(frame.image.get_obj())
            self.__subprocess.stdin.flush()

            fps(value=True)
        finally:
            frame.latch.next()

    def _close_stdin(self):
        try:
            self.__subprocess.stdin.close()
        except BrokenPipeError:
            pass  # Ignore broken pipe errors, as program exited before all data were written.

        if self.__written is None:
            # Interrupt waiting FFmpeg to let it finish stream without any data written to it.
            self.__subprocess.send_signal(signal.SIGINT)

    def start(self):
        super().start()
        self.__stderr_thread.start()
        if SIGCONT is not None:
            self.__subprocess.send_signal(signal.SIGCONT)

    def terminate(self):
        if SIGCONT is not None:
            self.__subprocess.send_signal(signal.SIGCONT)
        self.__subprocess.terminate()
        super().terminate()

    def join(self, timeout=None):
        try:
            super().join(timeout)
            self.__subprocess.wait(timeout)
        except Exception:
            self.__subprocess.terminate()
            raise
        finally:
            self.__subprocess.stderr.close()


def _stderr_reader(log_name, stream):
    wrapper = io.TextIOWrapper(stream)
    logger = logging.getLogger(log_name)
    logger.debug("Stderr redirected to stdout")
    try:
        line = wrapper.readline()
        while line:
            logger.info(line.rstrip())
            line = wrapper.readline()
    except Exception as e:
        logger.exception(e)
    finally:
        wrapper.close()
    logger.debug("Stderr gracefully closed")


class MpegTSReader(ReadPublish):
    """Reads encoded (compressed) video stream from FFmpeg subprocess by chunks equal to the size of
    a frame in the buffer. The buffer is filled with 188-byte sections of MPEG-TS stream, where each
    frame contains many of those packets, not necessarily the whole.
    Allows other workers to subscribe in order to re-stream the packets from the buffer.
    """

    def __init__(self, name: str, stop_event, log_queue, stream, frame_buffer, kwargs=None):
        super().__init__(name, stop_event, log_queue, stream, frame_buffer,
                         kwargs={} if kwargs is None else kwargs)

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super(Read, self)._run(stop_event, log_queue, *args, **kwargs)
        self._logger.debug("Reading from stdout")
        try:
            self._spin(self._process, stop_event, *args, **kwargs)
        except EOFError:
            pass  # Ignore when end of stream (file) reached and stop gracefully
        except Exception:
            self._logger.exception('Spin failure')
        finally:
            self._close(*args, **kwargs)
            self._logger.debug("Stdout closed")

    @staticmethod
    def _close(stream, *args, **kwargs):
        stream.close()

    def _new_frame(self, frame: Frame, stream, *args, **kwargs):
        frame.clear()
        image = stream.readinto(frame.image.get_obj())
        if not image:
            raise EOFError
        frame.header.epoch = time()
        return True



import cv2
import numpy as np
from queue import Empty
from threading import RLock
from watsor.stream.spin import Stub
from watsor.stream.work import Work, WorkInOutPublish, Payload
from watsor.stream.share import FrameBuffer, NonSharedFramesPerSecond


class VisualEffects(WorkInOutPublish):
    """Applies video effects after inference and filtering copying the entire frame from
    the input buffer to the output buffer. Allows to subscribe other workers to consume frames
    in the output buffer.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer_in, frame_buffer_out, effects,
                 kwargs=None):
        self.__fps = NonSharedFramesPerSecond()
        super().__init__(name, stop_event, log_queue, frame_queue, frame_buffer_in, frame_buffer_out,
                         args=(effects, self.__fps), kwargs={} if kwargs is None else kwargs)

    @property
    def fps(self):
        return self.__fps

    def _incoming_frame(self, frame_in, frame_out, stop_event, effects, fps, *args, **kwargs):
        try:
            image_shape, image_np_in = frame_in.get_numpy_image(np.uint8)
            _, image_np_out = frame_out.get_numpy_image(np.uint8)
            for e in effects:
                e.apply(image_np_in, image_np_out, image_shape, frame_in.header, frame_out.header)

            fps(value=True)
        finally:
            frame_in.latch.next()


class HttpStream(Work):
    """Base class for any other HTTP streamer.
    """

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer, subscriptions,
                 args=(), kwargs=None):
        self.__stop_event = stop_event
        self.__frame_queue = frame_queue
        self.__frame_buffer = frame_buffer
        self.__subscriptions = subscriptions
        self.__args = args
        self.__started = False
        super().__init__(Stub, name, stop_event, log_queue, frame_queue,
                         kwargs={} if kwargs is None else kwargs)

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super(Work, self)._run(stop_event, log_queue, *args, **kwargs)
        self._logger.debug("Started streaming")
        for consumer, topic in self.__subscriptions.items():
            consumer.subscribe(topic)
        self.__started = True

    def close(self):
        if self.__started:
            self.__started = False
            for consumer, topic in self.__subscriptions.items():
                consumer.unsubscribe(topic)
            self._deplete_queue(self.__frame_queue, self.__frame_buffer)
            self._logger.debug("Stopped streaming")

    @staticmethod
    def _deplete_queue(frame_queue, frame_buffer: FrameBuffer):
        """Read the queue till the end, returning frames into buffer
        """
        try:
            while True:
                payload = frame_queue.get_nowait()
                frame = frame_buffer.frames[payload.frame_index]
                frame.latch.next()
        except Empty:
            pass

    def __iter__(self):
        self.start()
        return self

    def __next__(self):
        if self.__stop_event.is_set():
            raise StopIteration()
        return self._process(self.__frame_queue, self.__stop_event, self.__frame_buffer, *self.__args)


class _JpegEncoderFrame(object):

    def __init__(self, lock=None):
        self.__lock = RLock() if lock is None else lock
        self.jpg = None
        self.epoch = 0

    @property
    def lock(self):
        return self.__lock


class MotionJpeg(HttpStream):

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer, encoder_buffer, subscriptions,
                 kwargs=None):
        super().__init__(name, stop_event, log_queue, frame_queue, frame_buffer, subscriptions,
                         args=(encoder_buffer,),
                         kwargs={} if kwargs is None else kwargs)

    @classmethod
    def create_buffer(cls, size):
        return [_JpegEncoderFrame() for _ in range(size)]

    @property
    def mime_type(self):
        return 'multipart/x-mixed-replace; boundary=--frame'

    def _next_frame(self, payload: Payload, stop_event, frame_buffer: FrameBuffer, encoder_buffer, *args,
                    **kwargs):
        frame_in = frame_buffer.frames[payload.frame_index]
        try:
            frame_out = encoder_buffer[payload.frame_index]
            with frame_out.lock:
                if frame_in.header.epoch == frame_out.epoch:
                    jpg = frame_out.jpg
                else:
                    image_shape, image_np = frame_in.get_numpy_image(np.uint8)
                    image = cv2.cvtColor(image_np, cv2.COLOR_BGR2RGB)
                    _, jpg = cv2.imencode('.jpg', image)

                    frame_out.jpg = jpg
                    frame_out.epoch = frame_in.header.epoch

            return self._part(jpg.tobytes())
        finally:
            frame_in.latch.next()

    def _no_frame(self, *args, **kwargs):
        return self._part(b'')

    @staticmethod
    def _part(jpg_bytes):
        msg = bytearray()
        msg.extend(b'--frame\r\n')
        msg.extend(b'Content-Type: image/jpeg\r\n')
        msg.extend(b'Content-Length: ' + f"{len(jpg_bytes)}".encode() + b'\r\n\r\n')
        msg.extend(jpg_bytes)
        msg.extend(b'\r\n')
        return bytes(msg)


class MpegTS(HttpStream):

    @property
    def mime_type(self):
        return 'video/mp2t'

    def _next_frame(self, payload: Payload, stop_event, frame_buffer: FrameBuffer, *args, **kwargs):
        frame = frame_buffer.frames[payload.frame_index]
        try:
            return bytes(frame.image.get_obj())
        finally:
            frame.latch.next()

    def _no_frame(self, *args, **kwargs):
        return b''



import json
import re
import paho.mqtt.client as mqtt
from distutils.util import strtobool
from time import time
from datetime import datetime
from threading import current_thread, RLock
from collections import namedtuple, defaultdict
from watsor.stream.work import Work, WorkPublish, Payload
from watsor.stream.share import FrameBuffer, FramesPerSecond
from watsor.config.coco import get_coco_class

UserData = namedtuple('userdata', ['topic', 'rate_limiter', 'stop_event'])


class ReportedState(object):

    def __init__(self):
        self.state = False
        self.when = 0.0


class NamedClient(mqtt.Client):

    def _thread_main(self):
        current_thread().name = self._client_id.decode('utf-8')
        super()._thread_main()

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port


# noinspection PyUnusedLocal
class MQTT(WorkPublish):
    """Communicates with HomeAssistant and others via MQTT.
    Reports state of the camera and its sensors such as FPS.
    Publishes binary state of a detected object, confirming the state every 10 seconds.
    Allows to start/stop the decoder, limit frame rate and enable/disable the reporting of detection details.

    List of topics:
    - watsor/cameras/{camera}/available
    - watsor/cameras/{camera}/command
    - watsor/cameras/{camera}/sensor
    - watsor/cameras/{camera}/state
    - watsor/cameras/{camera}/detection/{label}/state
    """

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, frame_buffer,
                 decoder_fps, decoder_rate_limiter, decoder_stop_event, mqtt_config, camera_config, kwargs=None):
        self.__fps = FramesPerSecond()
        self.__states = self.__init_states(camera_config)
        self.__sensors_hash = 0
        self.__old_state = None
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue, frame_buffer,
                         args=(mqtt_config, self.__fps, decoder_fps, decoder_rate_limiter, decoder_stop_event),
                         kwargs={} if kwargs is None else kwargs)

    @staticmethod
    def __init_states(camera_config):
        """Initialize the dictionary of detection states in order to start reporting OFF
        even if nothing's detected.

        :param camera_config: detection configuration for a camera
        :return: initialized default dictionary
        """
        result = defaultdict(ReportedState)
        for entry in camera_config['detect']:
            coco_class = next(iter(entry))
            _ = result[coco_class]
        return result

    @property
    def fps(self):
        return self.__fps

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super(Work, self)._run(stop_event, log_queue, *args, **kwargs)
        client, user_data = self._init_client(*args, **kwargs)
        if client is None:
            return
        try:
            self._init_locals()
            self._spin(self._process, stop_event, *args[:3], *args[4:], client, user_data, **kwargs)
        except Exception:
            self._logger.exception('Spin failure')
        finally:
            self._close_client(client, user_data)

    def _init_locals(self):
        self.__details = False
        self.__command_lock = RLock()
        self.__command_fps = re.compile(r"^fps\s*=\s*(\d+)$", re.IGNORECASE)
        self.__command_details = b = re.compile(r"^details\s*=\s*(\w+)$", re.IGNORECASE)

    def _init_client(self, frame_queue, stop_event, frame_buffer, config, fps, decoder_fps,
                     decoder_rate_limiter, decoder_stop_event, *args, **kwargs):
        try:
            client = NamedClient(client_id=self.name)

            if 'username' in config:
                client.username_pw_set(config['username'], config['password'])

            user_data = UserData(topic="{}/cameras/{}".format(kwargs['topic'], self.name),
                                 rate_limiter=decoder_rate_limiter,
                                 stop_event=decoder_stop_event)
            client.user_data_set(user_data)

            client.will_set(user_data.topic + '/available', payload='offline', qos=1, retain=True)

            client.on_log = self._on_log
            client.on_connect = self._on_connect
            client.on_disconnect = self._on_disconnect
            client.on_message = self._on_message

            client.connect(config['host'], config['port'])

            client.loop_start()
            return client, user_data
        except Exception as e:
            self._logger.error(e)
            return None, None

    def _close_client(self, client, user_data):
        try:
            client.loop_stop()
        except Exception:
            self._logger.exception('MQTT client failure')

    def _on_log(self, client, userdata, level, buf):
        if level < mqtt.MQTT_LOG_DEBUG:
            level_std = mqtt.LOGGING_LEVEL[level]
            self._logger.log(level_std, buf)

    def _on_connect(self, client, user_data, flags, rc):
        self._logger.debug("Connected to {}".format((client.host, client.port)) +
                           (" with result code {}".format(rc) if rc != 0 else ""))

        client.publish(user_data.topic + '/available', payload='online', qos=1, retain=True)

        client.subscribe(user_data.topic + '/command', qos=1)

    def _on_disconnect(self, client, user_data, rc):
        self._logger.debug("Disconnected from {}".format((client.host, client.port)) +
                           (" with result code {}".format(rc) if rc != 0 else ""))

    def _on_message(self, client, user_data, msg):
        """Allows to start/stop the decoder, limit frame rate and enable/disable the reporting of
         detection details.

        :param client: MQTT client
        :param user_data: the record with objects to control such as decoder stop event and rate limiter
        :param msg: protocol message (command)
        """
        command = str(msg.payload, 'utf-8')
        try:
            if command.upper() == 'ON':
                # Request to turn camera on
                if user_data.stop_event.is_set():
                    self._logger.debug("Turning camera on, wait for a while...")
                    user_data.stop_event.clear()
            elif command.upper() == 'OFF':
                # Request to turn camera off
                if not user_data.stop_event.is_set():
                    self._logger.debug("Turning camera off")
                    user_data.stop_event.set()
            else:
                match = self.__command_fps.match(command)
                if match:
                    # Request to limit FPS
                    rate = match.group(1)
                    user_data.rate_limiter.limit_rate(float(rate))
                    self._logger.debug("Limiting FPS to: {}".format(rate))
                    return

                match = self.__command_details.match(command)
                if match:
                    # Request for detection details
                    b = strtobool(match.group(1))
                    with self.__command_lock:
                        self.__details = b
                    self._logger.debug("Detection details: {}".format(b))
                    return

                raise ValueError("not recognized")
        except (AssertionError, ValueError) as e:
            self._logger.error("Invalid command '{}', {}".format(command, e))
        except Exception:
            self._logger.exception('MQTT client failure')

    def _new_frame(self, frame, payload: Payload, stop_event, frame_buffer: FrameBuffer,
                   fps, decoder_fps, decoder_rate_limiter, decoder_stop_event,
                   client, user_data, *args, **kwargs):
        try:
            groups = self._group_detections_by_label(frame.header.detections)

            self._publish_states_on(client, user_data, groups)
            self._publish_states_off(client, user_data, groups)
            self._publish_detections(client, user_data, groups, frame.header.epoch)
            self._publish_sensor_info(client, user_data, frame_buffer, fps(value=True), decoder_fps())
            self._publish_state(client, user_data)
        finally:
            frame.latch.next()

    def _no_frame(self, stop_event, frame_buffer: FrameBuffer,
                  fps, decoder_fps, decoder_rate_limiter, decoder_stop_event,
                  client, user_data, *args, **kwargs):
        """Even though a frame was not received, we need to update sensors and confirm state.
        """
        self._publish_sensor_info(client, user_data, frame_buffer, fps(), decoder_fps())
        self._publish_state(client, user_data)

    @staticmethod
    def _group_detections_by_label(detections):
        groups = defaultdict(list)
        for detection in filter(lambda d: d.label > 0, detections):
            label = get_coco_class(detection.label).label

            item = {'c': round(detection.confidence * 100, 1),
                    'b': [detection.bounding_box.x_min, detection.bounding_box.y_min,
                          detection.bounding_box.x_max, detection.bounding_box.y_max]}

            zones = [z for z in filter(lambda z: z > 0, detection.zones)]
            if len(zones):
                item['z'] = zones

            groups[label].append(item)
        return groups

    def _publish_states_on(self, client, user_data, groups):
        """Publish binary state of each detected label. The state should not retain as
        it can trigger alarm state after system pause/restart even the object is no longer
        present. Instead, the state is conformed/updated every 10 seconds.
        """
        now = time()
        for label, detections in groups.items():
            reported_state = self.__states[label]
            if not reported_state.state or (now - reported_state.when) >= 10:
                reported_state.state = True
                reported_state.when = now
                client.publish(topic="{}/detection/{}/state".format(user_data.topic, label), payload='ON', qos=1,
                               retain=False)

    def _publish_states_off(self, client, user_data, groups):
        """Turn OFF state of labels, which were ON, but no longer detected.
        """
        now = time()
        for label, reported_state in self.__states.items():
            if label not in groups:
                reported_state = self.__states[label]
                if reported_state.state or (now - reported_state.when) >= 10:
                    reported_state.state = False
                    reported_state.when = now
                    client.publish(topic="{}/detection/{}/state".format(user_data.topic, label), payload='OFF', qos=1,
                                   retain=False)

    def _publish_detections(self, client, user_data, groups, epoch):
        """Publish detection details.
        """
        with self.__command_lock:
            if not self.__details:
                return

        for label, detections in groups.items():
            details = {'t': datetime.fromtimestamp(epoch).isoformat(), 'd': detections}
            client.publish(topic="{}/detection/{}/details".format(user_data.topic, label),
                           payload=json.dumps(details))

    def _publish_sensor_info(self, client, user_data, frame_buffer, fps, decoder_fps):
        """Publish information about camera sensors, when values change.
        """
        sensor = {
            'fps_in': round(decoder_fps, 1),
            'fps_out': round(fps, 1),
            'buffer': round(frame_buffer.fullness * 100)
        }
        payload = json.dumps(sensor)
        payload_hash = payload.__hash__()
        if payload_hash != self.__sensors_hash:
            self.__sensors_hash = payload_hash
            client.publish(user_data.topic + '/sensor', payload, retain=True)

    def _publish_state(self, client, user_data):
        """Publish camera state: on/off.
        """
        new_state = 'OFF' if user_data.stop_event.is_set() else 'ON'
        if new_state != self.__old_state:
            self.__old_state = new_state
            client.publish(user_data.topic + '/state', payload=new_state, qos=1, retain=True)



import os
from tempfile import TemporaryDirectory, NamedTemporaryFile
from unittest import TestCase, main
from test.support import EnvironmentVarGuard
from uuid import uuid4
from watsor.config.loader import parse, validate, normalize, SECRETS_YAML


class TestConfig(TestCase):

    def test_empty_config(self):
        empty_yaml = """
        """

        with self.assertRaises(ValueError):
            validate(parse())
        with self.assertRaises(FileNotFoundError):
            validate(parse(filename=str(uuid4())))
        with self.assertRaises(AssertionError):
            validate(parse(data=empty_yaml))

    def test_required(self):
        minimal_config = """
        cameras:
            - porch:
        """

        with self.assertRaises(ValueError) as error:
            validate(parse(data=minimal_config))
        exception = error.exception

        self.assertRegex(exception.args[0], "cameras.0.porch.detect.*empty values not allowed")
        self.assertRegex(exception.args[0], "cameras.0.porch.height.*required field")
        self.assertRegex(exception.args[0], "cameras.0.porch.width.*required field")
        self.assertRegex(exception.args[0], "cameras.0.porch.input.*required field")
        self.assertRegex(exception.args[0], "cameras.0.porch.ffmpeg.decoder.*required field")

    def test_defaults(self):
        minimal_config = """
        ffmpeg:
            decoder: ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24']
            encoder: ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24']
        detect:
            - person:
        cameras:
            - porch:
                width: 640 
                height: 480
                input: http://192.168.42.129:8080/video
                mask: porch.png
                detect:
                    - person:
                    - truck: 
                        area: 80
                        confidence: 70           
        """

        config = normalize(validate(parse(data=minimal_config)), os.path.dirname(__file__))

        self.assertEqual(1, len(config['cameras']))
        camera = next(iter(config['cameras']))
        self.assertTrue('porch' in camera)
        self.assertEqual(640, camera['porch']['width'])
        self.assertEqual(480, camera['porch']['height'])
        self.assertEqual(8, len(camera['porch']['ffmpeg']['decoder']))
        self.assertEqual(10, len(camera['porch']['ffmpeg']['encoder']))
        detect = iter(camera['porch']['detect'])
        person = next(detect)
        self.assertTrue('person' in person)
        self.assertEqual(10, person['person']['area'])
        self.assertEqual(50, person['person']['confidence'])
        truck = next(detect)
        self.assertTrue('truck' in truck)
        self.assertEqual(80, truck['truck']['area'])
        self.assertEqual(70, truck['truck']['confidence'])

    def test_unique(self):
        minimal_config = """
        ffmpeg:
            decoder: ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24']
            encoder: ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24']
        detect:
            - person:
            - person:
        cameras:
            - porch:
                width: 640 
                height: 480
                input: http://192.168.42.129:8080/video
                detect:
                    - truck:
                    - truck:
            - porch:
                width: 640 
                height: 480
                input: http://192.168.42.129:8080/video
        """
        with self.assertRaises(ValueError) as error:
            validate(parse(data=minimal_config))
        exception = error.exception

        self.assertRegex(exception.args[0], "detect.1.*.'person' is already defined")
        self.assertRegex(exception.args[0], "cameras.1.*.'porch' is already defined")
        self.assertRegex(exception.args[0], "cameras.0.porch.detect.1.*'truck' is already defined")
        self.assertRegex(exception.args[0], "cameras.1.porch.detect.1.*'person' is already defined")

    def test_secrets(self):
        minimal_config = """
        mqtt:
            host: localhost
            username: !secret mqtt_username
            password: !secret mqtt_password
        ffmpeg:
            decoder: ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24']
            encoder: ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24']
        detect:
            - person:
        cameras:
            - porch:
                width: 640 
                height: 480
                input: http://192.168.42.129:8080/video
        """

        secrets = """
        mqtt_username: "john"
        mqtt_password: "qwerty"
        """

        with TemporaryDirectory() as tmp_dir:
            tmp_config_file = NamedTemporaryFile(mode='w', encoding="utf-8",
                                                 prefix=__name__, suffix=".yaml",
                                                 dir=tmp_dir, delete=False)
            try:
                tmp_config_file.write(minimal_config)
                tmp_config_file.flush()

                with open(os.path.join(tmp_dir, SECRETS_YAML),
                          'w', encoding="utf-8") as tmp_secrets_file:
                    tmp_secrets_file.write(secrets)
                    tmp_secrets_file.flush()

                    config = validate(parse(filename=tmp_config_file.name))
            finally:
                # Delete explicitly for compatibility with Windows NT
                tmp_config_file.close()
                os.unlink(tmp_config_file.name)

            self.assertEqual("john", config['mqtt']['username'])
            self.assertEqual("qwerty", config['mqtt']['password'])

    def test_env_vars(self):
        minimal_config = """
        http:
            port: 8080
            username: !env_var "ADMIN_USERNAME john"
            password: !env_var "ADMIN_PASSWORD qwerty"
        mqtt:
            host: localhost
            username: !ENV "${MQTT_ACCOUNT}@${MQTT_DOMAIN}"
            password: !ENV "${MQTT_PASSWORD}"
        cameras:
            - porch:
                width: 640 
                height: 480
                input: http://192.168.42.129:8080/video
                ffmpeg:
                    decoder: ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24']
                    encoder: ['-i', '-f', 'rawvideo', '-pix_fmt', 'rgb24']
                detect:
                    - person:
        """

        with EnvironmentVarGuard() as env:
            env.set("ADMIN_USERNAME", "admin")
            env.set("ADMIN_PASSWORD", "12345678")
            env.set("MQTT_ACCOUNT", "admin")
            env.set("MQTT_DOMAIN", "example.com")
            env.set("MQTT_PASSWORD", "qwerty")

            config = validate(parse(data=minimal_config))
            self.assertEqual("admin", config['http']['username'])
            self.assertEqual("12345678", config['http']['password'])
            self.assertEqual("admin@example.com", config['mqtt']['username'])
            self.assertEqual("qwerty", config['mqtt']['password'])


if __name__ == '__main__':
    main(verbosity=2)



import unittest
import multiprocessing as mp
import threading
import time
import itertools
from logging import getLogger
from logging.handlers import QueueHandler
from collections import defaultdict
from statistics import stdev
from watsor.stream.log import LogHandler
from watsor.stream.sync import BalancedQueue
from watsor.test.dummy_stream import DummyRead, DummyWork


class TestStream(unittest.TestCase):
    """Tests various configurations of streaming
    """

    def setUp(self):
        """Handle Ctrl+C signal to stop tests prematurely.
        """

        unittest.installHandler()

        self.stop_main_event = threading.Event()
        self.stop_process_event = mp.Event()

    def test_lag_absence(self):
        """Absence of lag during streaming, when the readers are much faster than workers
        Test configuration is set up so that the reader is several times faster then the worker.
        If buffering is configured, a chronic lag interferes the viewing of the processed stream in
        real time. The reader should drop the frames exceeded the buffer, but the buffer also should
        be small enough (up to 1 frame) to get rid of noticeable delay. If buffering is configured,
        the maximum lag between frames will match the reader's time: the frames are close to each
        other, but they are all in queue, which lags the viewing. Here we ensure the buffer is short
        and the maximum delay equals to worker's time, which means they are as up to date as possible.
        """

        # Prepare configuration
        distribution_queue = mp.Queue()
        log_handler, readers, workers, queues, semaphores = self.create_config(
            mp.Process, distribution_queue,
            read_config=[30],
            work_config=[10])

        # Run test for a second, then collect the distribution
        timeout = 1.05
        self.go(log_handler + readers + workers, timeout)

        distribution = queue_to_dict(distribution_queue)
        read_distr = select_and_strip(distribution, 'reader')
        work_distr = select_and_strip(distribution, 'worker')

        # Ensure the amount of data read and processed are the same
        self.assertEqual(sum(read_distr.values()), sum(work_distr.values()),
                         "The number of read and processed frames differ")

        # Ensure the maximum lag matches worker's time, not reader's time
        self.assertAlmostEqual(workers[0].max_lag, 1 / workers[0].fps, places=1,
                               msg="Maximum lag between frames differs from worker's time")

    def test_distribution_uniformity(self):
        """Uniform distribution, when capacity of readers well exceeds the workers
        Test configuration is set up so that total FPS capacity of the readers
        well exceeds FPS capacity of the workers. In this case the readers will compete
        for pushing their chunks of data to a slow worker. They won't be able to push all the
        data, but we need to ensure none of the readers is deprived and they all distributed
        almost equal number of frames. At that the workers will process the streams at their
        full capacity, which is configured different. We need to ensure there were no skews
        in workload.
        """

        # Prepare configuration
        distribution_queue = mp.Queue()
        log_handler, readers, workers, queues, semaphores = self.create_config(
            mp.Process, distribution_queue,
            read_config=[100, 300, 100, 300],
            work_config=[10, 50, 200])

        # Run test for a second, then collect the distribution
        timeout = 1.05
        self.go(log_handler + readers + workers, timeout)

        distribution = queue_to_dict(distribution_queue)
        read_distr = select_and_strip(distribution, 'reader')
        work_distr = select_and_strip(distribution, 'worker')

        # Ensure the amount of data read and processed are the same
        self.assertEqual(sum(read_distr.values()), sum(work_distr.values()),
                         "The number of read and processed frames differ")

        # Ensure the readers distribute chunks of data uniformly
        self.assertLess(stdev(read_distr.values()), 15,
                        "The readers did not distribute frames uniformly: {}"
                        .format(list(read_distr.values())))

        # Ensure the workload across the workers is distributed as configured
        for (i, j) in itertools.permutations(range(len(workers)), 2):
            self.assertAlmostEqual(first=workers[i].fps / workers[j].fps,
                                   second=work_distr[str(i)] / work_distr[str(j)],
                                   delta=5.0,
                                   msg="The workload per two workers ({} and {}) did not fit "
                                       "their configuration: planned [{}/{}], but was [{}/{}]"
                                   .format(i, j, workers[i].fps, workers[j].fps,
                                           work_distr[str(i)], work_distr[str(j)]))

    def test_idyll(self):
        """Idyll, when the workers can consume all data produced by the readers
        Test configuration is set up so that total FPS capacity of the workers
        well exceeds FPS capacity of the readers. In this case the readers will produce
        data at full capacity as the workers are able to process all frames.
        """

        # Prepare configuration
        distribution_queue = mp.Queue()
        log_handler, readers, workers, queues, semaphores = self.create_config(
            threading.Thread, distribution_queue,
            read_config=[30, 10, 30, 10],
            work_config=[160, 50])

        # Run test for a second, then collect the distribution
        timeout = 1.05
        timeout = self.go(log_handler + readers + workers, timeout)

        distribution = queue_to_dict(distribution_queue)
        read_distr = select_and_strip(distribution, 'reader')
        work_distr = select_and_strip(distribution, 'worker')

        # Ensure the amount of data read and processed are the same
        self.assertEqual(sum(read_distr.values()), sum(work_distr.values()),
                         "The number of read and processed frames differ")

        # Ensure each reader was used at full capacity
        for i in range(len(read_distr)):
            self.assertAlmostEqual(first=readers[i].fps,
                                   second=read_distr[str(i)] / timeout,
                                   delta=2.5,
                                   msg="The reader {} was not used at full capacity: planned {}, but was {}"
                                   .format(i, readers[i].fps, read_distr[str(i)] / timeout))

        # Ensure each reader's produced as much data as was configured, comparing with others
        for (i, j) in itertools.permutations(range(len(read_distr)), 2):
            self.assertAlmostEqual(first=readers[i].fps / readers[j].fps,
                                   second=read_distr[str(i)] / read_distr[str(j)],
                                   delta=5.0,
                                   msg="Two readers ({} and {}) produced diverse amount "
                                       "of data than was configured: planned [{}/{}], but was [{}/{}]"
                                   .format(i, j, readers[i].fps, readers[j].fps,
                                           read_distr[str(i)], read_distr[str(j)]))

    def create_config(self, delegate_class, distribution_queue, read_config, work_config):
        """Prepares configuration of a test.

        :param delegate_class: Process or Thread
        :param distribution_queue: queue to return distribution after test finishes
        :param read_config: FPS configuration of the readers
        :param work_config: FPS configuration of the workers
        :return: arrays of readers and workers
        """

        log_queue = mp.Queue()
        getLogger().addHandler(QueueHandler(log_queue))
        log_handler = LogHandler(delegate_class, "LogHandler", self.stop_process_event, log_queue, filename=None)

        frame_queue = mp.Queue()
        all_semaphores = {}

        readers = []
        for pos in range(len(read_config)):
            reader_name = "reader {}".format(pos)
            reader_queue_semaphore = mp.BoundedSemaphore(1)
            all_semaphores[reader_name] = reader_queue_semaphore
            decoder_queue = BalancedQueue(frame_queue, {reader_name: reader_queue_semaphore}, reader_name)

            readers.append(DummyRead(delegate_class, reader_name, self.stop_process_event, log_queue,
                                     decoder_queue, read_config[pos]))

        workers = []
        for pos in range(len(work_config)):
            workers.append(DummyWork(delegate_class, "worker {}".format(pos), self.stop_process_event, log_queue,
                                     BalancedQueue(frame_queue, all_semaphores),
                                     distribution_queue, work_config[pos]))

        return [log_handler], readers, workers, \
               [log_queue, frame_queue], all_semaphores

    def go(self, processes, timeout):
        """Starts all processes, waits till timeout exceeded,
        then terminates the processes.

        :param processes: list of processes to start
        :param timeout: timeout in seconds
        :return: time elapsed while waiting
        """

        for process in processes:
            process.start()

        timeout = self.wait(processes, timeout)

        self.stop_process_event.set()
        for process in processes:
            process.join(30)

        return timeout

    def wait(self, processes, timeout):
        """While waiting till timeout exceeded, checks each process is alive,
        stopping if not.

        :param processes: list of processes to check
        :param timeout: timeout in seconds
        :return: time elapsed while waiting
        """

        start_time = time.time()
        duration = 0
        while not self.stop_main_event.wait(timeout / 10) \
                and duration < timeout:
            if not all(map(lambda process: process.is_alive(), processes)):
                break
            duration = time.time() - start_time
        return time.time() - start_time


def queue_to_dict(queue):
    """Read all tuples from the queue, summing up the values for similar keys.

    :param queue: the queue to get tuples
    :return: dictionary with distribution data
    """

    result = defaultdict(int)
    while not queue.empty():
        (key, count) = queue.get()
        result[key] += count
    return result


def select_and_strip(dictionary, prefix):
    """Selects the keys from the dictionary by prefix,
    returning the filters dictionary with prefix stripped.

    :param dictionary: dictionary to filter
    :param prefix: prefix to match the keys
    :return: filtered dictionary
    """

    return {k[len(prefix):].strip(): v for k, v in dictionary.items() if k.startswith(prefix)}


if __name__ == '__main__':
    unittest.main(verbosity=2)



import cv2
import math
import numpy as np
import random
from time import time
from PIL import Image, ImageDraw
from collections import namedtuple
from watsor.stream.read import ReadDetectPublish
from watsor.stream.work import Work, WorkPublish, Payload
from watsor.stream.sync import CountDownLatch
from watsor.stream.share import Frame, FrameBuffer

Shape = namedtuple('Shape', ['index', 'bounding_box_area_ratio'])

SHAPE_TRIANGLE = Shape(1, 0.5)
SHAPE_ELLIPSE = Shape(2, 0.8)
SHAPE_RECTANGLE = Shape(3, 1.0)


class Artist(ReadDetectPublish):

    def __init__(self, name: str, stop_event, log_queue, frame_queue, frame_buffer):
        super().__init__(name, stop_event, log_queue, frame_queue, frame_buffer, args=(stop_event,))

    def _new_frame(self, frame: Frame, *args, **kwargs):
        frame.clear()

        # Create a new image and draw a shape
        with Image.new('RGB', (frame.header.width, frame.header.height)) as image:
            draw = ImageDraw.Draw(image)
            self.draw_random_shapes(image, draw)

            # Copy image to shared buffer
            flat_shape = (frame.header.height * frame.header.width * frame.header.channels,)
            image_np1 = np.array(image, copy=False).reshape(flat_shape)
            image_np2 = np.frombuffer(frame.image.get_obj(), np.uint8)
            np.copyto(image_np2, image_np1)

        frame.header.epoch = time()
        return True

    @classmethod
    def draw_random_shapes(cls, image, draw):
        width = image.width / 2
        height = image.height / 2
        center_x = image.width / 2
        center_y = image.height / 2
        cls._draw_random_shape(image, draw, 0, 0, width, height)
        cls._draw_random_shape(image, draw, center_x, 0, width, height)
        cls._draw_random_shape(image, draw, center_x, center_y, width, height)
        cls._draw_random_shape(image, draw, 0, center_y, width, height)

    @staticmethod
    def _draw_random_shape(image, draw, left, top, width, height):
        fill = (200, 200, 200)

        # Determine shape position
        shape = random.choice([SHAPE_ELLIPSE.index, SHAPE_RECTANGLE.index, SHAPE_TRIANGLE.index])
        p1_x = random.randrange(left, left + math.floor(width / 2))
        p1_y = random.randrange(top, top + math.floor(height / 2))
        p2_x = random.randrange(left + math.ceil(width / 2), left + width)
        p2_y = random.randrange(top + math.ceil(height / 2), top + height)

        # Draw one of three shapes
        if shape == SHAPE_ELLIPSE.index:
            draw.ellipse([(p1_x, p1_y), (p2_x, p2_y)], fill)
        elif shape == SHAPE_RECTANGLE.index:
            draw.rectangle([(p1_x, p1_y), (p2_x, p2_y)], fill)
        elif shape == SHAPE_TRIANGLE.index:
            draw.polygon([(p1_x, p1_y), (p1_x, p2_y), (p2_x, p2_y)], fill)


class ShapeDetector(Work):

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, frame_buffer):
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue,
                         args=(stop_event, frame_buffer))

    def _next_frame(self, payload: Payload, stop_event, frame_buffer: FrameBuffer):
        frame = frame_buffer.frames[payload.frame_index]
        try:
            image_shape, image_np = frame.get_numpy_image(np.uint8)

            # Convert to gray color and find contours
            gray_image = cv2.cvtColor(image_np, cv2.COLOR_BGR2GRAY)
            ret, thresh = cv2.threshold(gray_image, 100, 255, cv2.THRESH_BINARY)
            contours, _ = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[-2:]

            # Detect shape type and bounding box
            c, d = 0, 0
            while c < len(contours) and d < len(frame.header.detections):
                contour = contours[c]
                detection = frame.header.detections[d]
                if self._detect_shape(contour, detection):
                    d += 1
                c += 1
        finally:
            frame.latch.next()

    def _detect_shape(self, contour, detection):
        # Determine bounding box
        x, y, w, h = cv2.boundingRect(contour)
        if w <= 1 or h <= 1:
            return False

        detection.bounding_box.x_min = x
        detection.bounding_box.y_min = y
        detection.bounding_box.x_max = x + w - 1
        detection.bounding_box.y_max = y + h - 1

        # Detect shape type by comparing the shape area with the area of its bounding box
        ratio = cv2.contourArea(contour) / (w - 1) / (h - 1)

        if abs(ratio - SHAPE_TRIANGLE.bounding_box_area_ratio) <= 0.1:
            detection.confidence = self._calc_confidence(SHAPE_TRIANGLE, ratio)
            detection.label = SHAPE_TRIANGLE.index
        elif abs(ratio - SHAPE_ELLIPSE.bounding_box_area_ratio) <= 0.1:
            detection.confidence = self._calc_confidence(SHAPE_ELLIPSE, ratio)
            detection.label = SHAPE_ELLIPSE.index
        elif abs(ratio - SHAPE_RECTANGLE.bounding_box_area_ratio) <= 0.1:
            detection.confidence = self._calc_confidence(SHAPE_RECTANGLE, ratio)
            detection.label = SHAPE_RECTANGLE.index

        return True

    @staticmethod
    def _calc_confidence(shape, ratio):
        return 100 - abs(ratio - shape.bounding_box_area_ratio)


class ShapeCounter(WorkPublish):

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, frame_buffer, count_down_latch):
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue, frame_buffer,
                         args=(count_down_latch, ))

    def _new_frame(self, frame, payload: Payload, stop_event, frame_buffer: FrameBuffer, count_down_latch: CountDownLatch):
        try:
            for detection in frame.header.detections:
                if detection.label > 0:
                    count_down_latch.count_down()
        finally:
            frame.latch.next()



import numpy as np
from scipy.spatial import distance
from collections import defaultdict, deque
from watsor.stream.share import Detection


class TrackFilter(object):
    """Filters recurring detections being observed several times on a row. The parameter 'sensitivity'
    determines how many times a detection has to be observed to be filtered through. Helps to reduce
    false positives.

    This filter also levels flapping of the bounding box around a detected object. The flapping increases
    when different types of detectors process the same video stream. Due to the different object models,
    bounding box even around a stationary object moves constantly. The filter tracks the bounding boxes in
    adjacent frames, and groups together those where centroids are close, considering them belonging to the
    same object.
    """

    def __init__(self, filters=None, sensitivity=5, history=10):
        self.__sensitivity = sensitivity
        self.__history = history
        self.__filters = [] if filters is None else filters
        self.__detections_by_label = defaultdict(list)

    def __call__(self, detections):
        detections = filter(lambda d: d.label > 0 and all(f(d) for f in self.__filters), detections)
        return self._group_and_update(detections)

    def _group_and_update(self, detections):
        # Group detections by label
        groups = defaultdict(list)
        for detection in detections:
            groups[detection.label].append(detection)

        # When sensitivity is greater than 1, the result list of detections may be smaller
        # than the input list and the first findings won't be reported. However we want
        # to perform certain actions like resetting rate limit, even if it is not worth the report.
        suspicious_activity = len(groups) > 0

        # Remove labels that are no longer detected
        for label in list(self.__detections_by_label.keys()):
            if label not in groups:
                for detection_history in self.__detections_by_label[label]:
                    detection_history.clear()
                self.__detections_by_label[label].clear()
                del self.__detections_by_label[label]

        for label, detections in groups.items():
            # Calculate centroids for each input bounding box
            len_bounding_boxes = len(detections)
            input_centroids = np.zeros((len_bounding_boxes, 2), dtype="int")
            for i, detection in enumerate(detections):
                input_centroids[i] = self._centroid(detection.bounding_box)

            # Calculate centroids for each previously found bounding box
            len_existing_centroids = len(self.__detections_by_label[label])
            existing_centroids = np.zeros((len_existing_centroids, 2), dtype="int")
            for i, detection_history in enumerate(self.__detections_by_label[label]):
                existing_centroids[i] = self._centroid(detection_history[0].bounding_box)

            # Compute the distance between each pair of object centroids and input centroids
            distances = distance.cdist(np.array(existing_centroids), input_centroids)
            if len(distances.shape) == 2 and distances.shape[0] > 0 and distances.shape[1] > 0:
                # Find the smallest value in each row and then
                # sort the row indexes in ascending order of their values
                rows = np.argsort(np.amin(distances, axis=1))
                # Find the smallest value in each column and then
                # sort using the previously computed row index list
                cols = np.argmin(distances, axis=1)[rows]
            else:
                rows = []
                cols = []

            # Consider bounding boxes, which centroids are close, belonging to the same object,
            # keep them together.
            used_rows = set()
            used_cols = set()
            for (row, col) in zip(rows, cols):
                if row in used_rows or col in used_cols:
                    continue

                self.__detections_by_label[label][row].append(detections[col])

                used_rows.add(row)
                used_cols.add(col)

            unused_rows = set(range(len_existing_centroids)).difference(used_rows)
            unused_cols = set(range(len_bounding_boxes)).difference(used_cols)

            # Remove bounding boxes that are no longer detected
            for row in sorted(unused_rows, reverse=True):
                self.__detections_by_label[label][row].clear()
                del self.__detections_by_label[label][row]

            # Add bounding boxes that were not present
            for col in unused_cols:
                history = deque([detections[col]], maxlen=self.__history)
                self.__detections_by_label[label].append(history)

        # For each detection we return a maximum bounding box of all its bounding boxes
        # found within last 10 frames to avoid flapping
        result = []
        for label, detections in self.__detections_by_label.items():
            for detection_history in detections:
                # Skip detection if it's observed less than required number of times
                if len(detection_history) < self.__sensitivity:
                    continue

                detection = self._combine(detection_history)
                result.append(detection)
        return result, suspicious_activity

    @staticmethod
    def _centroid(bounding_box):
        cx = int((bounding_box.x_min + bounding_box.x_max) / 2.0)
        cy = int((bounding_box.y_min + bounding_box.y_max) / 2.0)
        return cx, cy

    @staticmethod
    def _combine(detection_history):
        new_detection = Detection()
        new_detection.label = detection_history[0].label
        new_detection.confidence = detection_history[0].confidence
        new_detection.bounding_box.x_min = detection_history[0].bounding_box.x_min
        new_detection.bounding_box.y_min = detection_history[0].bounding_box.y_min
        new_detection.bounding_box.x_max = detection_history[0].bounding_box.x_max
        new_detection.bounding_box.y_max = detection_history[0].bounding_box.y_max

        for i in range(1, len(detection_history)):
            detection = detection_history[i]
            new_detection.confidence = max(new_detection.confidence, detection.confidence)
            new_detection.bounding_box.x_min = min(new_detection.bounding_box.x_min, detection.bounding_box.x_min)
            new_detection.bounding_box.y_min = min(new_detection.bounding_box.y_min, detection.bounding_box.y_min)
            new_detection.bounding_box.x_max = max(new_detection.bounding_box.x_max, detection.bounding_box.x_max)
            new_detection.bounding_box.y_max = max(new_detection.bounding_box.y_max, detection.bounding_box.y_max)

        zones = set()
        for detection in detection_history:
            for zone in filter(lambda z: z > 0, detection.zones):
                zones.add(zone)

        iterator = iter(zones)
        for zone in range(len(new_detection.zones)):
            try:
                new_detection.zones[zone] = next(iterator)
            except StopIteration:
                new_detection.zones[zone] = 0

        return new_detection



import numpy as np
from collections import namedtuple

CocoClass = namedtuple('CocoClass', [
    'label',
    'box_color',
    'font_color',
    'box_thickness',
    'font_thickness',
    'font_scale',
    'alpha',
])

COCO_CLASSES = [
    'unlabeled',
    'person',
    'bicycle',
    'car',
    'motorcycle',
    'airplane',
    'bus',
    'train',
    'truck',
    'boat',
    'traffic light',
    'fire hydrant',
    'street sign',
    'stop sign',
    'parking meter',
    'bench',
    'bird',
    'cat',
    'dog',
    'horse',
    'sheep',
    'cow',
    'elephant',
    'bear',
    'zebra',
    'giraffe',
    'hat',
    'backpack',
    'umbrella',
    'shoe',
    'eye glasses',
    'handbag',
    'tie',
    'suitcase',
    'frisbee',
    'skis',
    'snowboard',
    'sports ball',
    'kite',
    'baseball bat',
    'baseball glove',
    'skateboard',
    'surfboard',
    'tennis racket',
    'bottle',
    'plate',
    'wine glass',
    'cup',
    'fork',
    'knife',
    'spoon',
    'bowl',
    'banana',
    'apple',
    'sandwich',
    'orange',
    'broccoli',
    'carrot',
    'hot dog',
    'pizza',
    'donut',
    'cake',
    'chair',
    'couch',
    'potted plant',
    'bed',
    'mirror',
    'dining table',
    'window',
    'desk',
    'toilet',
    'door',
    'tv',
    'laptop',
    'mouse',
    'remote',
    'keyboard',
    'cell phone',
    'microwave',
    'oven',
    'toaster',
    'sink',
    'refrigerator',
    'blender',
    'book',
    'clock',
    'vase',
    'scissors',
    'teddy bear',
    'hair drier',
    'toothbrush',
]

# Random RGB colors for each label
_COCO_COLORS = np.random.RandomState(255).uniform(0, 255, size=(len(COCO_CLASSES), 3)).astype(np.uint8)

# Dictionary of labels with index as a key
_COCO_DICTIONARY = {
    idx: CocoClass(label,
                   box_color=tuple(map(int, _COCO_COLORS[idx])),
                   font_color=(255, 255, 255),
                   box_thickness=1,
                   font_thickness=1,
                   font_scale=0.5,
                   alpha=0.55)
    for idx, label in enumerate(COCO_CLASSES)
}


def get_coco_class(idx):
    """Finds COCO class in build-in dictionary, returning 'unlabeled' if not found.

    :param idx: index of COCO class to find
    :return: CocoClass instance
    """

    return _COCO_DICTIONARY.get(idx, _COCO_DICTIONARY[0])



from io import DEFAULT_BUFFER_SIZE
from watsor.stream.spin import Spin


class Copy(Spin):
    """Copies data from one stream to another.
    """

    def __init__(self, delegate_class, name: str, stop_event, log_queue, src, dst,
                 buffer_size=DEFAULT_BUFFER_SIZE, args=(), kwargs=None):
        super().__init__(delegate_class, name, stop_event, log_queue,
                         args=(src, dst, buffer_size, *args), kwargs={} if kwargs is None else kwargs)

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super()._run(stop_event, log_queue, *args, **kwargs)
        try:
            self._spin(self._copy, stop_event, *args, **kwargs)
        except BrokenPipeError:
            pass  # Ignore broken pipe errors, as program exited before all data were written
        except Exception:
            self._logger.exception('Spin failure')
        finally:
            self._close(*args, **kwargs)

    @staticmethod
    def _copy(src, dst, buffer_size, *args, **kwargs):
        buf = src.read(buffer_size)
        if buf:
            dst.write(buf)

    @staticmethod
    def _close(src, dst, *args, **kwargs):
        try:
            dst.close()
        except OSError:
            pass  # Ignore OS errors such as broken pipe, as program most likely exited before all data were written
        finally:
            src.close()



import os
import re
import yaml
import cerberus
import logging
from watsor.config.schema import schema

_LOGGER = logging.getLogger(__name__)

_ENV_pattern = re.compile('.*?\${(\w+)}.*?')

SECRETS_YAML = "secrets.yaml"

__SECRET_CACHE = {}


def _load_yaml(filename: str, loader: yaml.Loader = yaml.SafeLoader):
    _LOGGER.debug("Loading %s", filename)
    with open(filename, encoding="utf-8") as stream:
        return yaml.load(stream, Loader=loader)


def _env_yaml(loader: yaml.Loader, node: yaml.nodes.Node):
    """Extracts the environment variable from the node's value

    :param loader: the yaml loader
    :param node: the current node in the yaml
    :return: the parsed string that contains the value of the environment variable
    """

    value = loader.construct_scalar(node)
    match = _ENV_pattern.findall(value)  # to find all env variables in line
    if match:
        full_value = value
        for g in match:
            full_value = full_value.replace(
                f'${{{g}}}', os.environ.get(g, g)
            )
        return full_value
    return value


def _env_var_yaml(loader: yaml.Loader, node: yaml.nodes.Node) -> str:
    """Load environment variables and embed it into the configuration YAML.

    :param loader: the yaml loader
    :param node: the current node in the yaml
    :return: the parsed string that contains the value of the environment variable
    """

    args = node.value.split()

    # Check for a default value
    if len(args) > 1:
        return os.getenv(args[0], " ".join(args[1:]))
    if args[0] in os.environ:
        return os.environ[args[0]]
    raise ValueError(node.value)


def _load_secret_yaml(filename: str):
    """Load the secrets yaml from file.

    :param filename: the name of the secret file
    :return: the contents of the secret file as a dictionary
    """

    if filename in __SECRET_CACHE:
        return __SECRET_CACHE[filename]

    try:
        secrets = _load_yaml(filename)
        if secrets is None:
            raise FileNotFoundError()
        if not isinstance(secrets, dict):
            raise ValueError("Secrets is not a dictionary")
    except FileNotFoundError:
        secrets = {}

    __SECRET_CACHE[filename] = secrets
    return secrets


def _secret_yaml(loader: yaml.Loader, node: yaml.nodes.Node):
    """Load secrets and embed it into the configuration YAML.

    :param loader: the yaml loader
    :param node: the current node in the yaml
    :return: the parsed string that contains the value of the secret
    """

    secret_path = os.path.dirname(loader.name)
    while True:
        filename = os.path.join(secret_path, SECRETS_YAML)
        secrets = _load_secret_yaml(filename)

        if node.value in secrets:
            _LOGGER.debug("Secret \"%s\" retrieved from %s", node.value, filename)
            return secrets[node.value]

        secret_path = os.path.dirname(secret_path)
        if not os.path.exists(secret_path) or len(secret_path) < 5:
            break  # Somehow we got past the config folder

    raise ValueError(f"Secret \"{node.value}\" not defined")


def parse(filename=None, data=None):
    """
    Load a yaml configuration file and resolve any environment variables and secrets.

    :param str filename: the path to the yaml file
    :param str data: the yaml data of file is not provided
    :return: the dict configuration or None if file is empty
    :rtype: dict[str, object]
    """

    loader: yaml.Loader = yaml.SafeLoader

    loader.add_implicit_resolver('!ENV', _ENV_pattern, None)
    loader.add_constructor('!ENV', _env_yaml)
    loader.add_constructor("!env_var", _env_var_yaml)
    loader.add_constructor("!secret", _secret_yaml)

    __SECRET_CACHE.clear()
    try:
        if filename:
            return _load_yaml(filename, loader=loader)
        elif data:
            return yaml.load(data, Loader=loader)
        else:
            raise ValueError('Either filename or data should be defined as input')
    finally:
        __SECRET_CACHE.clear()


class _ExtendedValidator(cerberus.Validator):
    """Performs project-specific normalization of configuration:
    - copies ffmpeg and detect defaults to the settings of a camera;
    - checks uniqueness of detection labels and camera names in Yaml lists.
    """

    def _normalize_default_setter_ffmpeg(self, _):
        return self.root_document['ffmpeg']

    def _normalize_default_setter_detect(self, _):
        return self.root_document['detect']

    def _validate_uniquekey(self, uniquekey, field, value):
        """Ensure value uniqueness.
        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if uniquekey:
            before = [self.document[i] for i in range(0, field)]
            this_key = next(iter(value))
            before_keys = filter(lambda item: item == this_key,
                                 map(lambda doc: next(iter(doc)), before))
            count = sum(1 for _ in before_keys)
            if count > 0:
                msg = "'{}' is already defined".format(this_key)
                self._error(field, msg)


def _walk_dict(a_dict, visitor, path=()):
    """Walks through a dictionary tree (where a leaf can be also a dictionary)
    without recursion.

    :param a_dict: a dictionary tree to walk through
    :param visitor: callable to invoke for each non-dictionary leaf found
    :param path: start path
    """
    stack = [(a_dict, path)]
    while stack:
        dict_node, path_node = stack.pop(0)
        for key, value in dict_node.items():
            next_path_node = path_node + (str(key),)
            for item in value:
                if isinstance(item, dict):
                    stack.append((item, next_path_node))
                else:
                    visitor(next_path_node, item)


def validate(config):
    """
    Validate configuration against embedded schema.

    :param config: configuration object
    :raises: AssertionError if config is None, ValueError if validation failed
    """

    assert config is not None, "Configuration file is empty"
    validator = _ExtendedValidator(schema)

    if not validator.validate(config, schema):
        str_list = ["Invalid configuration:"]
        _walk_dict(validator.errors, lambda paths, error: str_list
                   .append("\t\"{}\": {}"
                           .format(".".join(paths), error)))
        raise ValueError("\n".join(str_list))

    return validator.normalized(config)


def normalize(config, path):
    """Normalizes config inserting input and output parameters in FFmpeg command line.
    Resolves relative paths in mask file.

    :param config: config object
    :param path: path to the file, where config was loaded from
    :return: normalized config object
    """

    for camera in config['cameras']:
        camera_name = next(iter(camera))
        camera_config = camera[camera_name]
        ffmpeg = camera_config['ffmpeg']

        item = 'decoder'
        ffmpeg[item].insert(0, 'ffmpeg')
        input_index = ffmpeg[item].index('-i')
        ffmpeg[item].insert(input_index + 1, camera_config['input'])
        ffmpeg[item].append('-')

        item = 'encoder'
        if item in ffmpeg:
            ffmpeg[item].insert(0, 'ffmpeg')
            input_index = ffmpeg[item].index('-i')
            size = '{}x{}'.format(camera_config['width'], camera_config['height'])
            ffmpeg[item].insert(input_index, '-s')
            ffmpeg[item].insert(input_index + 1, size)
            ffmpeg[item].insert(input_index + 3, '-')
            if 'output' in camera_config:
                ffmpeg[item].append(camera_config['output'])
            else:
                ffmpeg[item].append('-')

        if 'mask' in camera_config:
            mask = camera_config['mask']
            if not os.path.isabs(mask):
                camera_config['mask'] = os.path.realpath(os.path.join(path, mask))

    return config



import os
import cv2
import numpy as np
import pycuda.driver as cuda
import tensorrt as trt
import watsor.engine as engine
from collections import namedtuple
from operator import itemgetter
from pycuda.tools import clear_context_caches
from typing import List
from time import time
from watsor.stream.share import Detection

HostDeviceMem = namedtuple('HostDeviceMem', 'host device')


class TensorRTObjectDetector:
    """Performs object detection on Nvidia CUDA GPUs.
    TensorRT takes a trained TensorFlow network and produces a highly optimized runtime engine.
    """

    def __init__(self, model_path, device):
        cuda.init()
        self.__device = cuda.Device(device)
        self.context = self.__device.make_context()

        trt.init_libnvinfer_plugins(engine.TRT_LOGGER, '')

        self.__trt_runtime = trt.Runtime(engine.TRT_LOGGER)
        try:
            self.__trt_engine = engine.load_engine(self.__trt_runtime, os.path.join(model_path, 'gpu.buf'))
        except Exception as e:
            self.__finalize()
            raise e

        self._allocate_buffers()

        self.__model_shape = itemgetter(1, 2)(self.__trt_engine.get_binding_shape('Input'))
        self.__execution_context = self.__trt_engine.create_execution_context()

    def __finalize(self):
        self.context.pop()
        self.context = None

        clear_context_caches()

    @property
    def device_name(self):
        return self.__device.name()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__finalize()

    def detect(self, image_shape, image_np, detections: List[Detection]):
        # Resize to model shape
        image_tensor = cv2.resize(image_np, dsize=self.__model_shape, interpolation=cv2.INTER_LINEAR)
        # HWC -> CHW
        image_tensor = image_tensor.transpose((2, 0, 1))
        # Normalize to [-1.0, 1.0] interval (expected by model)
        image_tensor = (2.0 / 255.0) * image_tensor - 1.0
        image_tensor = image_tensor.ravel()

        np.copyto(self.inputs[0].host, image_tensor)
        inference_start_time = time()
        detection_out, keep_count_out = self._do_inference()
        inference_time = (time() - inference_start_time) * 1000

        d = 0
        max_width = image_shape[1] - 1
        max_height = image_shape[0] - 1
        while d < int(keep_count_out[0]) and d < len(detections):
            detection = detections[d]
            pred_start_idx = d * 7
            detection.label = detection_out[pred_start_idx + 1]
            detection.confidence = detection_out[pred_start_idx + 2]
            detection.bounding_box.x_min = int(detection_out[pred_start_idx + 3] * max_width)
            detection.bounding_box.y_min = int(detection_out[pred_start_idx + 4] * max_height)
            detection.bounding_box.x_max = int(detection_out[pred_start_idx + 5] * max_width)
            detection.bounding_box.y_max = int(detection_out[pred_start_idx + 6] * max_height)
            d += 1

        return inference_time

    def _do_inference(self):
        [cuda.memcpy_htod_async(inp.device, inp.host, self.stream) for inp in self.inputs]
        self.__execution_context.execute_async(batch_size=self.__trt_engine.max_batch_size,
                                               bindings=self.bindings,
                                               stream_handle=self.stream.handle)
        [cuda.memcpy_dtoh_async(out.host, out.device, self.stream) for out in self.outputs]
        self.stream.synchronize()
        return [out.host for out in self.outputs]

    def _allocate_buffers(self):
        self.inputs = []
        self.outputs = []
        self.bindings = []
        self.stream = cuda.Stream()

        # NMS implementation in TRT 6 only supports DataType.FLOAT
        binding_to_type = {"Input": np.float32,
                           "NMS": np.float32,
                           "NMS_1": np.int32,
                           "Postprocessor": np.float32,
                           "Postprocessor_1": np.int32}
        for binding in self.__trt_engine:
            shape = self.__trt_engine.get_binding_shape(binding)
            size = trt.volume(shape) * self.__trt_engine.max_batch_size
            dtype = binding_to_type[str(binding)]

            # Allocate host and device buffers
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)

            # Append the device buffer to device bindings.
            self.bindings.append(int(device_mem))

            # Append to the appropriate list.
            if self.__trt_engine.binding_is_input(binding):
                self.inputs.append(HostDeviceMem(host_mem, device_mem))
            else:
                self.outputs.append(HostDeviceMem(host_mem, device_mem))



import argparse
import os

try:
    import tensorrt as trt
except ImportError as e:
    if __name__ != '__main__':
        raise e
    else:
        print('TensorRT is not installed, skipping.')
        exit()

TRT_LOGGER = trt.Logger(trt.Logger.WARNING)


def build_engine(uff_model_path, trt_engine_datatype=trt.DataType.FLOAT,
                 batch_size=1, model_width=300, model_height=300):
    with trt.Builder(TRT_LOGGER) as builder, \
            builder.create_network() as network, \
            trt.UffParser() as uff_parser:
        builder.max_workspace_size = 1 << 30
        builder.max_batch_size = batch_size
        if trt_engine_datatype == trt.DataType.HALF:
            builder.fp16_mode = True

        uff_parser.register_input("Input", (3, model_width, model_height))
        uff_parser.register_output("MarkOutput_0")
        uff_parser.parse(uff_model_path, network)

        return builder.build_cuda_engine(network)


def save_engine(engine, engine_dest_path):
    os.makedirs(os.path.dirname(engine_dest_path), exist_ok=True)
    buf = engine.serialize()
    with open(engine_dest_path, 'wb') as f:
        f.write(buf)


def load_engine(trt_runtime, engine_path):
    with open(engine_path, 'rb') as f:
        engine_data = f.read()
    engine = trt_runtime.deserialize_cuda_engine(engine_data)
    return engine


TRT_PRECISION_TO_DATATYPE = {
    16: trt.DataType.HALF,
    32: trt.DataType.FLOAT
}

if __name__ == '__main__':
    # Define script command line arguments
    parser = argparse.ArgumentParser(description='Utility to build TensorRT engine prior to inference.')
    parser.add_argument('-i', "--input",
                        dest='uff_model_path', metavar='UFF_MODEL_PATH', required=True,
                        help='preprocessed TensorFlow model in UFF format')
    parser.add_argument('-p', '--precision', type=int, choices=[32, 16], default=32,
                        help='desired TensorRT float precision to build an engine with')
    parser.add_argument('-b', '--batch-size', type=int, default=1,
                        help='max TensorRT engine batch size')
    parser.add_argument('-mw', '--model-width', type=int, default=300,
                        help='SSD model width')
    parser.add_argument('-mh', '--model-height', type=int, default=300,
                        help='SSD model height')
    parser.add_argument("-o", "--output", dest='trt_engine_path',
                        help="path of the output file",
                        default=os.path.join(os.path.dirname(os.path.abspath(__file__)), "engine.buf"))

    # Parse arguments passed
    args = parser.parse_args()

    # Build TensorRT engine
    print("Building TensorRT engine. This may take few minutes.")
    trt.init_libnvinfer_plugins(TRT_LOGGER, '')
    trt_engine = build_engine(
        uff_model_path=args.uff_model_path,
        trt_engine_datatype=TRT_PRECISION_TO_DATATYPE[args.precision],
        batch_size=args.batch_size,
        model_width=args.model_width, model_height=args.model_height)
    if trt_engine is not None:
        # Save the engine to file
        save_engine(trt_engine, args.trt_engine_path)
        print("TensorRT engine saved to {}".format(args.trt_engine_path))



import random
import numpy as np
from time import time
from collections import namedtuple
from multiprocessing import Value
from watsor.stream.read import Read
from watsor.stream.work import Work
from watsor.stream.share import FrameBuffer

Action = namedtuple('Action', ['frame_index', 'operation', 'value', 'expected_result'])

OPERATIONS = "*/+-"


def simple_math(operation, x, y):
    return {
        OPERATIONS[0]: lambda _x: _x * y,
        OPERATIONS[1]: lambda _x: _x / y,
        OPERATIONS[2]: lambda _x: _x + y,
        OPERATIONS[3]: lambda _x: _x - y
    }[operation](x)


class NumpyRead(Read):

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, frame_buffer):
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue,
                         args=(stop_event, frame_buffer))
        self._last_frame_index = -1

    def _next_frame(self, frame_queue, stop_event, frame_buffer: FrameBuffer):
        frame, frame_index = frame_buffer.select_next_ready(self._last_frame_index)
        self._last_frame_index = frame_index
        if frame is None:
            return None

        image_shape, image_np = frame.get_numpy_image()

        # Fill image buffer with random data
        data = np.random.randn(*image_shape)
        np.copyto(image_np, data)

        # Calculate sum of all pixels and do math against the sum
        x = np.sum(image_np[:]).item()
        y = random.random()
        operation = OPERATIONS[random.randint(0, len(OPERATIONS[0]))]
        expected_result = simple_math(operation, x, y)

        frame.header.epoch = time()
        frame.latch.next()

        return Action(frame_index, operation, y, expected_result)


class NumpyWork(Work):

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, frame_buffer):
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue,
                         args=(stop_event, frame_buffer))
        self._matches = Value('i', 0)

    def _next_frame(self, action: Action, stop_event, frame_buffer: FrameBuffer):
        frame = frame_buffer.frames[action.frame_index]
        try:
            # Get image numpy buffer
            image_shape = (frame.header.height, frame.header.width, frame.header.channels)
            image_np = np.frombuffer(frame.image.get_obj()).reshape(image_shape)

            # Do math against numpy image, changing the raw data
            simple_math(action.operation, image_np, action.value)

            # Calculate sum of all pixels and compare with expected value
            x = np.sum(image_np[:]).item()
            result = simple_math(action.operation, x, action.value)
            if round(result - action.expected_result, 1) == 0:
                # Count position match
                self._matches.value = self._matches.value + 1
        finally:
            frame.latch.next()

    @property
    def matches(self):
        return self._matches.value



import os
import cv2
import numpy as np
import tflite_runtime.interpreter as tflite
from operator import itemgetter
from typing import List
from time import time
from watsor.stream.share import Detection


class TensorFlowLiteObjectDetector:
    """TensorFlow Lite inference on CPU is used on IoT devices.
    TensorFlow Lite running on desktop without any GPU delegate is 9 times worse than TensorFlow.
    """

    def __init__(self, model_path):
        self.__interpreter = tflite.Interpreter(model_path=os.path.join(model_path, 'cpu.tflite'))
        self.__interpreter.allocate_tensors()

        self.__tensor_input_details = self.__interpreter.get_input_details()
        self.__tensor_output_details = self.__interpreter.get_output_details()

        self.__model_shape = itemgetter(1, 2)(self.__tensor_input_details[0]['shape'])

    @property
    def device_name(self):
        return "CPU"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def detect(self, image_shape, image_np, detections: List[Detection]):
        image_np = cv2.resize(image_np, dsize=self.__model_shape, interpolation=cv2.INTER_LINEAR)

        inference_start_time = time()
        self.__interpreter.set_tensor(self.__tensor_input_details[0]['index'],
                                      np.expand_dims(image_np, axis=0))
        self.__interpreter.invoke()

        boxes = np.squeeze(self.__interpreter.get_tensor(self.__tensor_output_details[0]['index']))
        label_codes = np.squeeze(self.__interpreter.get_tensor(self.__tensor_output_details[1]['index'])) + 1
        scores = np.squeeze(self.__interpreter.get_tensor(self.__tensor_output_details[2]['index']))

        inference_time = (time() - inference_start_time) * 1000

        d = 0
        max_width = image_shape[1] - 1
        max_height = image_shape[0] - 1
        while d < len(scores) and d < len(detections):
            detection = detections[d]
            detection.label = label_codes[d]
            detection.confidence = scores[d]
            detection.bounding_box.y_min = min(int(boxes[d][0] * max_height), max_height)
            detection.bounding_box.x_min = min(int(boxes[d][1] * max_width), max_width)
            detection.bounding_box.y_max = min(int(boxes[d][2] * max_height), max_height)
            detection.bounding_box.x_max = min(int(boxes[d][3] * max_width), max_width)
            d += 1

        return inference_time



import os
import subprocess
from multiprocessing import set_start_method
from argparse import ArgumentParser

if __name__ == '__main__':
    set_start_method('spawn')

    parser = ArgumentParser(description='Object detection for video surveillance')
    parser.add_argument("--model-path",
                        dest='model_path', metavar='MODEL_PATH',
                        default=os.path.join(os.getcwd(), 'model'),
                        help="path to log file")

    args, unknown = parser.parse_known_args()

    if os.path.isfile(os.path.join(args.model_path, 'gpu.uff')) and not \
            os.path.isfile(os.path.join(args.model_path, 'gpu.buf')):
        engine = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'engine.py')
        subprocess.run(['python3', '-u', engine,
                        '-i', os.path.join(args.model_path, 'gpu.uff'),
                        '-o', os.path.join(args.model_path, 'gpu.buf'),
                        '-p', os.getenv('TRT_FLOAT_PRECISION', '32')
                        ], check=True)

    from watsor.main import Application

    app = Application()
    app.run()



from threading import Thread
from watsor.stream.spin import Spin


class WatchDog(Spin):
    """Watches the other processes, checking every 10 seconds and restarting them if not alive.
    """

    def __init__(self, name: str, stop_event, log_queue, interval=10, kwargs=None):
        super().__init__(Thread, name, stop_event, log_queue,
                         args=(interval,), kwargs={} if kwargs is None else kwargs)

    __children = []

    @classmethod
    def add_child(cls, child):
        cls.__children.append(child)

    @classmethod
    def remove_child(cls, child):
        cls.__children.remove(child)

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super()._run(stop_event, log_queue, *args, **kwargs)
        self._logger.debug("Started")
        try:
            self._spin(self._watch, stop_event, stop_event, *args, **kwargs)
        except Exception:
            self._logger.exception('Spin failure')
        self._logger.debug("Stopped")

    def _watch(self, stop_event, interval, *args, **kwargs):
        for child in self.__children:
            if child.is_alive():
                self._logger.debug('{} {} ({}) is alive'.format(child.delegate_class_name,
                                                                child.name, child.__class__.__name__))
            elif not child.is_shutdown():
                self._restart(child)

        stop_event.wait(interval)

    def _restart(self, child):
        try:
            self._logger.warning('{} {} ({}) is not alive, restarting...'.format(
                child.delegate_class_name, child.name, child.__class__.__name__))
            child.initialize()
            child.start()
        except AssertionError as e:
            self._logger.error('Failed to restart {} {} ({}). {}.'.format(
                child.delegate_class_name, child.name, child.__class__.__name__, e))
        except Exception:
            self._logger.exception('Failed to restart {} {} ({})'.format(
                child.delegate_class_name, child.name, child.__class__.__name__))



import unittest
from logging import getLogger, ERROR
from logging.handlers import QueueHandler
from threading import Thread
from multiprocessing import Event, Queue
from watsor.stream.log import LogHandler
from watsor.stream.watch import WatchDog
from watsor.stream.sync import CountDownLatch
from watsor.test.dummy_stream import Stumble


class TestWatch(unittest.TestCase):

    def setUp(self):
        """Handle Ctrl+C signal to stop tests prematurely.
        """

        unittest.installHandler()

    def test_watch_dog(self):
        """Runs the watchdog and a child thread that stumbles all the time,
        forcing the watchdog to restart it. Counts the number of restarts during timeout.
        """

        log_queue = Queue()
        stop_process_event = Event()

        getLogger().addHandler(QueueHandler(log_queue))

        latch = CountDownLatch(3)

        log_handler = LogHandler(Thread, "logger", stop_process_event, log_queue, filename=None)
        watch_dog = WatchDog("watchdog", stop_process_event, log_queue, 0.1, kwargs={'log_level': ERROR})
        stumble = Stumble(Thread, "tumbler", stop_process_event, log_queue, 0.1, latch)

        log_handler.start()
        watch_dog.start()
        stumble.start()

        watch_dog.add_child(log_handler)
        watch_dog.add_child(stumble)

        self.assertTrue(latch.wait(5))
        stop_process_event.set()

        watch_dog.remove_child(log_handler)
        watch_dog.remove_child(stumble)

        stumble.join(30)
        watch_dog.join(30)
        log_handler.join(30)


if __name__ == '__main__':
    unittest.main(verbosity=2)



import cv2
from shapely.geometry import Polygon
from watsor.stream.share import Detection
from watsor.config.coco import COCO_CLASSES


class MaskFilter(object):
    """Filters only detections where bounding box intersects the opaque area in mask image.
    Mask image must be of 32 bit color and of the same resolution as corresponding camera
    video stream. Detection zone are selected in alpha channel of a mask image.

    All zones in mask image are sorted and indexed depending on the proximity of their center
    to the origin. When the bounding box detected intersects a zone, the index of that zone
    is marked in detection record.
    """

    def __init__(self, camera_config):
        # Find contours in alpha channel
        filename = camera_config['mask']
        alpha_channel, _ = get_alpha_channel(filename,
                                             camera_config['width'],
                                             camera_config['height'])
        contours = find_contours(alpha_channel)

        # Convert contours to polygons
        self.__polygons = [Polygon(c[:, 0]) for c in contours]

        # Prepare zones dictionary
        self.__polygons_by_zone = {}
        for entry in camera_config['detect']:
            coco_class = next(iter(entry))
            index = COCO_CLASSES.index(coco_class)

            zones = entry[coco_class]['zones']
            if len(zones) == 0:
                continue
            for z in zones:
                assert 0 < z <= len(self.__polygons), \
                    "There is no zone {} in mask {}".format(z, filename)

            self.__polygons_by_zone[index] = [p if idx + 1 in zones else None
                                              for idx, p in enumerate(self.__polygons)]

    def __call__(self, detection: Detection):
        bounding_box = Polygon([(detection.bounding_box.x_min, detection.bounding_box.y_min),
                                (detection.bounding_box.x_max, detection.bounding_box.y_min),
                                (detection.bounding_box.x_max, detection.bounding_box.y_max),
                                (detection.bounding_box.x_min, detection.bounding_box.y_max)])
        polygons = self.__polygons_by_zone.get(detection.label, self.__polygons)
        result = False
        z = 0
        p = 0
        while p < len(polygons) and z < len(detection.zones):
            if polygons[p] is not None and bounding_box.intersects(polygons[p]):
                detection.zones[z] = p + 1
                z += 1
                result = True
            p += 1
        return result


def get_alpha_channel(filename, width=None, height=None):
    mask_image = cv2.imread(filename, cv2.IMREAD_UNCHANGED)

    assert mask_image is not None, \
        "Error reading mask file {}".format(filename)

    assert len(mask_image.shape) == 3 and mask_image.shape[2] == 4, \
        "Mask image {} is not of 32 bit color".format(filename)

    if width is not None and height is not None:
        assert mask_image.shape[0] == height and mask_image.shape[1] == width, \
            "The size of mask image {} doesn't match {}x{}".format(filename, width, height)

    return mask_image[:, :, 3], mask_image


def contours_key(contour):
    moments = cv2.moments(contour)
    center = (int(moments['m10'] / moments['m00']), int(moments['m01'] / moments['m00']))
    return center[0] * center[0] + center[1] * center[1]


def find_contours(alpha_channel):
    ret, thresh = cv2.threshold(255 - alpha_channel, 0, 255, cv2.THRESH_BINARY_INV)
    contours, _ = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[-2:]
    contours = sorted(contours, key=contours_key)
    return contours



import os
from unittest import TestCase, installHandler, main
from test.support import EnvironmentVarGuard
from threading import Thread
from logging import getLogger
from logging.handlers import QueueHandler
from multiprocessing import Process, Event, Queue
from watsor.stream.log import LogHandler
from watsor.stream.share import FrameBuffer
from watsor.stream.sync import CountDownLatch
from watsor.detection.detector import create_object_detectors
from watsor.filter.sieve import DetectionSieve
from watsor.filter.track import TrackFilter
from watsor.filter.confidence import ConfidenceFilter
from watsor.stream.share import RateLimiter
from watsor.config.coco import get_coco_class
from watsor.test.detect_stream import Artist, ShapeCounter


class TestDetect(TestCase):

    def setUp(self):
        if not os.path.exists(self._get_model_path()):
            self.skipTest('Test skipped due to the absence of shape object model')

        installHandler()

    def test_shape_detection(self):
        """Tests TensorFlow object detection using the trained model of simple geometric shapes.
        The detector recognises shapes drawn on a frame image, the sieve filters those having confidence
        above 50%, lastly shapes are counted signalling to end the test.
        """

        with EnvironmentVarGuard() as env:
            env.set("TF_CPP_MIN_LOG_LEVEL", "3")
            env.set("CORAL_VISIBLE_DEVICES", "")
            env.set("CUDA_VISIBLE_DEVICES", "")

            frame_buffer = FrameBuffer(10, 100, 100)

            frame_queue = Queue(1)
            subscriber_queue = Queue(1)
            detection_sieve_queue = Queue(1)

            log_queue = Queue()
            getLogger().addHandler(QueueHandler(log_queue))

            stop_process_event = Event()

            latch = CountDownLatch(100)

            artist = Artist("artist", stop_process_event, log_queue, frame_queue, frame_buffer)

            detection_sieve = DetectionSieve("sieve", stop_process_event, log_queue,
                                             detection_sieve_queue, frame_buffer,
                                             self._create_filters(), RateLimiter())

            processes = [artist, detection_sieve,
                         LogHandler(Thread, "logger", stop_process_event, log_queue, filename=None),
                         ShapeCounter(Thread, "counter", stop_process_event, log_queue, subscriber_queue,
                                      frame_buffer, latch)]

            processes += create_object_detectors(Process, stop_process_event, log_queue, frame_queue,
                                                 {artist.name: frame_buffer}, self._get_model_path())

            artist.subscribe(detection_sieve_queue)
            detection_sieve.subscribe(subscriber_queue)

            for process in processes:
                process.start()

            try:
                self.assertTrue(latch.wait(15))
            finally:
                stop_process_event.set()
                for process in processes:
                    process.join(30)

    @staticmethod
    def _create_filters():
        return [TrackFilter([ConfidenceFilter({
            'detect': [{
                get_coco_class(1).label: {  # triangle
                    'confidence': 50
                }}, {
                get_coco_class(2).label: {  # ellipse
                    'confidence': 50
                }}, {
                get_coco_class(3).label: {  # rectangle
                    'confidence': 50
                }}]})], 1, 1)]

    @staticmethod
    def _get_model_path():
        return os.path.join(os.path.dirname(__file__), 'model')


if __name__ == '__main__':
    main(verbosity=2)



from time import time
from collections import namedtuple, defaultdict
from multiprocessing import Value
from watsor.stream.spin import Spin
from watsor.stream.read import Read
from watsor.stream.work import Work

Frame = namedtuple('Frame', ['device', 'time'])


class DummyRead(Read):

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, fps):
        self._fps = fps
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue,
                         args=(stop_event,))

    def _next_frame(self, frame_queue, stop_event):
        if stop_event.wait(1 / self._fps):
            return None

        return Frame(self.name, time())

    @property
    def fps(self):
        return self._fps


class DummyWork(Work):

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue, done_queue, fps):
        self._fps = fps
        self._last_frame_time = float('inf')
        self._max_lag = Value('d', 0)
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue,
                         args=(stop_event, self._max_lag),
                         kwargs={'done_queue': done_queue})

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super(Work, self)._run(stop_event, log_queue, *args, **kwargs)
        try:
            count = defaultdict(int)
            self._spin(self._process, stop_event, *args, count)
            self._done(count, **kwargs)
        except Exception:
            self._logger.exception('Run failure')

    def _next_frame(self, frame: Frame, stop_event, max_lag, count):
        max_lag.value = max(frame.time - self._last_frame_time, max_lag.value)
        self._last_frame_time = frame.time

        count[frame.device] += 1  # Count the number of frames produced by each reader

        stop_event.wait(1 / self._fps)

    def _done(self, count, done_queue):
        # Report the readers distribution data
        for key in count:
            done_queue.put_nowait((key, count[key]))
        # Report the number of frames processed by this worker
        done_queue.put_nowait((self.name, sum(count.values())))

    @property
    def fps(self):
        return self._fps

    @property
    def max_lag(self):
        return self._max_lag.value


class Stumble(Spin):

    def __init__(self, delegate_class, name: str, stop_event, log_queue, interval, count_down_latch, kwargs=None):
        super().__init__(delegate_class, name, stop_event, log_queue,
                         args=(interval, count_down_latch),
                         kwargs={} if kwargs is None else kwargs)

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super()._run(stop_event, log_queue, *args, **kwargs)
        self._wait(stop_event, *args, **kwargs)
        self._logger.debug("Exited")

    def _wait(self, stop_event, interval, count_down_latch, *args, **kwargs):
        count_down_latch.count_down()
        self._logger.debug("Started, waiting for {} sec".format(interval))
        stop_event.wait(interval)



from os import path
from numpy import uint8
from typing import Dict
from multiprocessing.sharedctypes import Array
from watsor.stream.work import Work, Payload
from watsor.stream.share import FrameBuffer, FramesPerSecond, InferenceTime
from watsor.detection.devices import cuda_gpus, edge_tpus, cpus

_ALWAYS_USE_CPU = False


def create_object_detectors(delegate_class, stop_event, log_queue, frame_queue, frame_buffers, model_path, kwargs=None):
    """Creates all available detectors.
    Falls back to CPU if neither GPU or Coral detected.

    :param delegate_class: either Thread of Process
    :param stop_event: event to stop the detection process
    :param log_queue: queue to put log messages
    :param frame_queue: queue to wait for a frame to process
    :param frame_buffers: dictionary of frame buffers. A frame payload contains
                          `sender` key to find a buffer in the dictionary,
    :param model_path: path where the NN model is
    :param kwargs: dictionary with options such as log level
    :return: list of detector processes to start
    """

    detectors = []
    if kwargs is None:
        kwargs = {}

    def gen_name():
        return "detector{}".format(len(detectors) + 1)

    def append_detector(_clazz, *_args):
        detectors.append(ObjectDetector(delegate_class, gen_name(), stop_event, log_queue, frame_queue, frame_buffers,
                                        kwargs={**kwargs,
                                                'detector_class': _clazz,
                                                'detector_args': (model_path, *_args,)}))

    if path.isfile(path.join(model_path, 'edgetpu.tflite')):
        for device, clazz in edge_tpus():
            append_detector(clazz, device)

    if path.isfile(path.join(model_path, 'gpu.buf')):
        for device, clazz in cuda_gpus():
            append_detector(clazz, device)

    if _ALWAYS_USE_CPU or len(detectors) == 0:
        for clazz in cpus():
            append_detector(clazz)

    assert len(detectors) > 0, "Failed to create an object detector." \
                               "Make sure TensorFlow is installed and model files are provided."

    return detectors


class ObjectDetector(Work):
    """Performs object detection inference without modifying image data,
    but filling detection records.
    """

    def __init__(self, delegate_class, name: str, stop_event, log_queue, frame_queue_in,
                 frame_buffers: Dict[object, FrameBuffer], kwargs=None):
        self.__fps = FramesPerSecond()
        self.__inference_time = InferenceTime()
        self.__device_name = Array('c', 255)
        super().__init__(delegate_class, name, stop_event, log_queue, frame_queue_in,
                         args=(stop_event, frame_buffers, self.__fps, self.__inference_time),
                         kwargs={} if kwargs is None else kwargs)

    @property
    def device_name(self):
        return self.__device_name.value

    @property
    def fps(self):
        return self.__fps

    @property
    def inference_time(self):
        return self.__inference_time

    def _run(self, stop_event, log_queue, *args, **kwargs):
        super(Work, self)._run(stop_event, log_queue, *args, **kwargs)
        try:
            detector_class = kwargs.get('detector_class')
            detector_args = kwargs.get('detector_args')

            with detector_class(*detector_args) as object_detector:
                self.__device_name.value = str.encode(object_detector.device_name)[:len(self.__device_name)]

                self._logger.debug("{}{} initialized".format(
                    object_detector.__class__.__name__, detector_args))

                self._spin(self._process, stop_event, *args, object_detector, **kwargs)
        except FileNotFoundError as e:
            self._logger.error(e)
        except Exception:
            self._logger.exception('Detection failure')

    def _next_frame(self, payload: Payload, stop_event, frame_buffers, fps, inference_time,
                    object_detector, *args, **kwargs):
        frame = frame_buffers[payload.sender].frames[payload.frame_index]
        try:
            image_shape, image_np = frame.get_numpy_image(uint8)
            time_of_inference = object_detector.detect(image_shape, image_np, frame.header.detections)

            inference_time(value=time_of_inference)
            fps(value=True)
        finally:
            frame.latch.next()
